import hashlib
import logging
from pathlib import Path

from cg.constants import FileExtensions

LOG = logging.getLogger(__name__)
BYTES_PER_CHUNK: int = 8192


def check_md5sum(file_path: Path, md5sum: str) -> bool:
    """Checks if the given md5_sum matches that of the given file"""
    with open(file_path, "rb") as file:
        file_hash = hashlib.md5()
        chunk = file.read(BYTES_PER_CHUNK)
        while chunk:
            file_hash.update(chunk)
            chunk = file.read(BYTES_PER_CHUNK)
    calculated_md5sum = file_hash.hexdigest()
    if md5sum == calculated_md5sum:
        return True
    LOG.info(f"The given md5sum does not match the md5sum for file {file_path.as_posix()}")
    return False


def extract_md5sum(md5sum_file: Path) -> str:
    """Searches the first line of a file for a string matching the length of an md5sum"""
    with open(md5sum_file, "r") as md_file:
        first_line: str = md_file.readline()
    line_list = first_line.split(" ")
    for element in line_list:
        element = element.strip()
        if len(element) == 32 and "." not in element:
            md5sum = element
            return md5sum
    LOG.info(f"No valid md5sum found in file {md5sum_file.as_posix()}")
    return ""


def is_md5sum_correct(file_path: Path) -> bool:
    """Return True if the given file exists and matches its md5sum."""
    if Path(str(file_path) + FileExtensions.MD5).exists():
        given_md5sum: str = extract_md5sum(md5sum_file=Path(str(file_path) + FileExtensions.MD5))
        return check_md5sum(file_path=file_path, md5sum=given_md5sum)
    return False



from cg.services.orders.validation.models.order_with_cases import OrderWithCases
from cg.services.orders.validation.order_types.nallo.constants import NalloDeliveryType
from cg.services.orders.validation.order_types.nallo.models.case import NalloCase
from cg.services.orders.validation.order_types.nallo.models.sample import NalloSample


class NalloOrder(OrderWithCases[NalloCase]):
    delivery_type: NalloDeliveryType

    @property
    def enumerated_new_samples(self) -> list[tuple[int, int, NalloSample]]:
        return [
            (case_index, sample_index, sample)
            for case_index, case in self.enumerated_new_cases
            for sample_index, sample in case.enumerated_new_samples
        ]



from cg.services.orders.validation.models.order_with_cases import OrderWithCases
from cg.services.orders.validation.order_types.balsamic_umi.constants import BalsamicUmiDeliveryType
from cg.services.orders.validation.order_types.balsamic_umi.models.case import BalsamicUmiCase


class BalsamicUmiOrder(OrderWithCases[BalsamicUmiCase]):
    delivery_type: BalsamicUmiDeliveryType



from http import HTTPStatus

from flask import Blueprint, Response, request

from cg.exc import AnalysisDoesNotExistError, TrailblazerAPIHTTPError
from cg.server.endpoints.utils import before_request
from cg.server.ext import db, mark_as_delivered_service
from cg.store.models import Analysis

DELIVER_BLUEPRINT = Blueprint("deliver", __name__, url_prefix="/api/v1")
DELIVER_BLUEPRINT.before_request(before_request)


@DELIVER_BLUEPRINT.route("/deliver", methods=["POST"])
def deliver_analyses():
    """
    Marks the analyses as delivered.
    - Adds a date stamp to the 'delivered_at' field for the relevant samples.
    - Calls Trailblazer to mark the Trailblazer analyses entries as delivered.
    """
    try:
        trailblazer_ids: list[int] = request.json["trailblazer_ids"]  # type: ignore None is handled
        analyses = []
        for trailblazer_id in trailblazer_ids:
            analysis: Analysis = db.get_analysis_by_trailblazer_id(trailblazer_id)
            analyses.append(analysis)
        mark_as_delivered_service.mark_analyses(analyses)
    except (AnalysisDoesNotExistError, KeyError, TypeError):
        return Response(status=HTTPStatus.BAD_REQUEST)
    except TrailblazerAPIHTTPError:
        db.rollback()
        return Response(status=HTTPStatus.BAD_GATEWAY)
    finally:
        db.commit_to_store()
    return Response(status=HTTPStatus.NO_CONTENT)



"""Code for uploading coverage reports via CLI"""

import rich_click as click

from cg.apps.coverage import ChanjoAPI, chanjo_api_for_genome_build
from cg.constants.constants import GenomeBuild
from cg.meta.upload.coverage import UploadCoverageApi
from cg.models.cg_config import CGConfig
from cg.store.models import Case
from cg.store.store import Store

from .utils import suggest_cases_to_upload


@click.command("coverage")
@click.argument("family_id", required=False)
@click.option(
    "--genome-version",
    type=click.Choice(["hg19", "hg38"]),
    default="hg19",
    help="Which chanjo instance to upload to",
)
@click.pass_obj
def upload_coverage(context: CGConfig, family_id: str | None, genome_version: str):
    """Upload coverage from an analysis to Chanjo."""

    click.echo(click.style("----------------- COVERAGE --------------------"))

    status_db: Store = context.status_db

    if not family_id:
        suggest_cases_to_upload(status_db=status_db)
        raise click.Abort

    case: Case = status_db.get_case_by_internal_id(internal_id=family_id)
    chanjo_api: ChanjoAPI = chanjo_api_for_genome_build(
        config=context, genome_build=GenomeBuild[genome_version]
    )
    upload_coverage_api = UploadCoverageApi(
        status_api=status_db,
        hk_api=context.housekeeper_api,
        chanjo_api=chanjo_api,
    )
    coverage_data = upload_coverage_api.data(case.latest_completed_analysis)
    upload_coverage_api.upload(coverage_data)



import logging
from datetime import datetime

from cg.constants import DataDelivery, Workflow
from cg.models.orders.sample_base import StatusEnum
from cg.services.orders.constants import ORDER_TYPE_WORKFLOW_MAP
from cg.services.orders.lims_service.service import OrderLimsService
from cg.services.orders.storing.service import StoreOrderService
from cg.services.orders.validation.order_types.pacbio_long_read.models.order import PacbioOrder
from cg.services.orders.validation.order_types.pacbio_long_read.models.sample import PacbioSample
from cg.store.models import ApplicationVersion, Case, CaseSample, Customer, Order, Sample
from cg.store.store import Store

LOG = logging.getLogger(__name__)


class StorePacBioOrderService(StoreOrderService):
    """Storing service for PacBio Long Read orders."""

    def __init__(self, status_db: Store, lims_service: OrderLimsService):
        self.status_db = status_db
        self.lims = lims_service

    def store_order(self, order: PacbioOrder) -> dict:
        """Store the order in the statusDB and LIMS, return the database samples and LIMS info."""
        project_data, lims_map = self.lims.process_lims(
            samples=order.samples,
            ticket=order._generated_ticket_id,
            order_name=order.name,
            workflow=Workflow.RAW_DATA,
            customer=order.customer,
            delivery_type=DataDelivery(order.delivery_type),
            skip_reception_control=order.skip_reception_control,
        )
        self._fill_in_sample_ids(samples=order.samples, lims_map=lims_map)
        new_samples = self.store_order_data_in_status_db(order=order)
        return {"project": project_data, "records": new_samples}

    def store_order_data_in_status_db(self, order: PacbioOrder) -> list[Sample]:
        """
        Store all order data in the Status database for a Pacbio order. Return the samples.
        The stored data objects are:
        - Order
        - Samples
        - For each Sample, a Case
        - For each Sample, a relationship between the Sample and its Case
        """
        status_db_order: Order = self._create_db_order(order=order)
        new_samples = []
        with self.status_db.no_autoflush_context():
            for sample in order.samples:
                case: Case = self._create_db_case_for_sample(
                    sample=sample,
                    customer=status_db_order.customer,
                    order=order,
                )
                db_sample: Sample = self._create_db_sample(
                    sample=sample,
                    order_name=order.name,
                    customer=status_db_order.customer,
                    ticket_id=str(status_db_order.ticket_id),
                )
                case_sample: CaseSample = self.status_db.relate_sample(
                    case=case, sample=db_sample, status=StatusEnum.unknown
                )
                self.status_db.add_multiple_items_to_store([case, case_sample, db_sample])
                status_db_order.cases.append(case)
                new_samples.append(db_sample)
        self.status_db.add_item_to_store(status_db_order)
        self.status_db.commit_to_store()
        return new_samples

    def _create_db_order(self, order: PacbioOrder) -> Order:
        """Return an Order database object."""
        ticket_id: int = order._generated_ticket_id
        customer: Customer = self.status_db.get_customer_by_internal_id(
            customer_internal_id=order.customer
        )
        return self.status_db.add_order(customer=customer, ticket_id=ticket_id)

    def _create_db_case_for_sample(
        self, sample: PacbioSample, customer: Customer, order: PacbioOrder
    ) -> Case:
        """Return a Case database object for a PacbioSample."""
        case_name: str = f"{sample.name}-case"
        case: Case = self.status_db.add_case(
            data_analysis=ORDER_TYPE_WORKFLOW_MAP[order.order_type],
            data_delivery=DataDelivery(order.delivery_type),
            name=case_name,
            priority=sample.priority,
            ticket=str(order._generated_ticket_id),
        )
        case.customer = customer
        return case

    def _create_db_sample(
        self, sample: PacbioSample, order_name: str, customer: Customer, ticket_id: str
    ) -> Sample:
        """Return a Sample database object."""
        application_version: ApplicationVersion = (
            self.status_db.get_current_application_version_by_tag(tag=sample.application)
        )
        return self.status_db.add_sample(
            name=sample.name,
            customer=customer,
            application_version=application_version,
            sex=sample.sex,
            comment=sample.comment,
            internal_id=sample._generated_lims_id,
            order=order_name,
            ordered=datetime.now(),
            original_ticket=ticket_id,
            priority=sample.priority,
            tumour=sample.tumour,
        )
